package net.pingex.discordbot;

import org.apache.commons.lang3.concurrent.BasicThreadFactory;
import sx.blah.discord.api.EventSubscriber;
import sx.blah.discord.handle.impl.events.MessageReceivedEvent;
import sx.blah.discord.util.DiscordException;
import sx.blah.discord.util.HTTP429Exception;
import sx.blah.discord.util.MissingPermissionsException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.concurrent.*;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This class parses and dispatch Discord commands
 * @version 0.1-dev
 * @author Raphael "Pingex" NAAS
 */
class CommandDispatcher
{
	private Logger logger;

	/**
	 * Singleton instance
	 */
	private static final CommandDispatcher INSTANCE = new CommandDispatcher();

	/**
	 * Contains all available commands, built using `rebuildCommandList()`
	 */
	private HashMap<String, InvokableMethod> commandList;

	/**
	 * Contains all the shorthanded commands
	 */
	private HashMap<String, String> shortList;

	/**
	 * The command prefix used to call commands
	 */
	private String commandPrefix;

	/**
	 * Thread pool used to invoke commands
	 */
	private ExecutorService threadPool;

	/**
	 * Basic Constructor, automatically rebuilds command list from the modules registry
	 */
	private CommandDispatcher()
	{
		logger = Logger.getLogger(this.getClass().getName());
		rebuildCommandList();
		threadPool = Executors.newCachedThreadPool(new BasicThreadFactory.Builder().namingPattern("CommandDispatcher-%d").build()); // TODO: threadPool.shutdown()

		if(Configuration.exists("general", "commandPrefix") && !Configuration.getValue("general", "commandPrefix").isEmpty())
			commandPrefix = Configuration.getValue("general", "commandPrefix");
		else
		{
			logger.warning("Invalid command prefix detected, falling back to \"!\"");
			commandPrefix = "!";
		}
	}

	/**
	 * Gets the unique instance of `CommandDispatcher`
	 * @return The current instance of `CommandDispatcher`
	 */
	public static CommandDispatcher getInstance()
	{
		return INSTANCE;
	}

	/**
	 * Fired when receiving a message
	 * @param event Event from Discord API
	 */
	@EventSubscriber
	public void onMessageReceivedEvent(MessageReceivedEvent event)
	{
		// Command matcher
		Matcher m = Pattern.compile("^" + commandPrefix + "([:\\w]+)(?: (.*))?$").matcher(event.getMessage().getContent());
		if(!m.matches()) return; // We don't need to go further if it's not even a command
		String fullCommand = m.group(1);

		// Arg splitter
		Matcher am = Pattern.compile("([^\"]\\S*|\".+?\")\\s*").matcher((m.group(2) != null) ? m.group(2) : "");
		ArrayList<String> args = new ArrayList<>();
		while(am.find()) args.add(am.group(1).replace("\"", ""));

		// Shorthand matcher
		if(!fullCommand.contains(":") && shortList.containsKey(fullCommand))
			fullCommand = shortList.get(fullCommand);

		String commandAnswer = null;
		Object[] parsedArray = null;

		// Conditions
		try
		{
			if(commandList.containsKey(fullCommand))
			{
				logger.info("Command invoked (" + event.getMessage().getAuthor().getName() + "#" + event.getMessage().getAuthor().getDiscriminator() + "): " + fullCommand);

				InvokableMethod foundMethod = commandList.get(fullCommand);
				parsedArray = new Object[foundMethod.getMethod().getParameterCount()];
				parsedArray[0] = event;

				if(foundMethod.getMethod().getParameterCount()-1 == args.size())
				{
					for(int i=1; i < foundMethod.getMethod().getParameterCount(); i++)
						try
						{
							parsedArray[i] = parse(foundMethod.getMethod().getParameterTypes()[i], args.get(i-1));
						} catch (IllegalArgumentException e)
						{
							commandAnswer = "Failed to parse arguments, are they correct ? " + commandList.get("internal:help").invoke(event, fullCommand);
							break;
						}
				}
				else
					commandAnswer = "Invalid arguments. " + commandList.get("internal:help").invoke(event, fullCommand);
			}
			else
				commandAnswer = "Unknown command";
		}
		catch (InvocationTargetException | IllegalAccessException e)
		{
			logger.severe("Couldn't get help: " + e.getMessage());
		}

		// Run command
		if(commandAnswer == null)
		{
			String finalFullCommand = fullCommand;
			Object[] finalParsedArray = parsedArray;
			threadPool.submit(() -> {
				try
				{
					String ans = (String) commandList.get(finalFullCommand).invoke(finalParsedArray);
					if(ans != null) event.getMessage().reply(ans);
				} catch (IllegalArgumentException | IllegalAccessException | InvocationTargetException e)
				{
					logger.severe("Couldn't call target method (" + e.getClass().getName() + "): " + e.getMessage());
				} catch (MissingPermissionsException | HTTP429Exception | DiscordException e)
				{
					logger.warning("Couldn't reply to command (" + e.getClass().getName() + "): " + e.getMessage());
				} catch (Exception e)
				{
					logger.severe("Error in threaded command");
					e.printStackTrace();
				}
			});
		}
		else // Error answer
			try
			{
				event.getMessage().reply(commandAnswer);
			} catch (MissingPermissionsException | HTTP429Exception | DiscordException e)
			{
				logger.warning("Couldn't reply to command (" + e.getClass().getName() + "): " + e.getMessage());
			}
	}

	/**
	 * Rebuilds command list using the module registry.
	 * - It looks for methods with @Command annotation inside class with @Controllable annotation
	 * - Check his prototype (should be `method(IDiscordClient, ...):String`)
	 * - Add usable methods in his local registry, using InvokableMethod
	 * @see InvokableMethod
	 * @see ModulesRegistry
	 */
	public void rebuildCommandList()
	{
		logger.info("Rebuilding command list...");

		ArrayList<AbstractModule> registry = ModulesRegistry.getRegistry();
		commandList = new HashMap<>();
		shortList = new HashMap<>();

		for(AbstractModule i : registry)
			if(i.getClass().isAnnotationPresent(Controllable.class))
				for(Method j : i.getClass().getDeclaredMethods())
					if(j.isAnnotationPresent(Command.class))
					{
						String id = i.getClass().getAnnotation(Controllable.class).name() + ":" + j.getName();
						logger.info("Found " + id);

						if(j.getParameterCount() >= 1 && j.getParameterTypes()[0] == MessageReceivedEvent.class && j.getReturnType() == String.class)
							if(!commandList.containsKey(id))
							{
								commandList.put(id, new InvokableMethod(j, i));

								// Eventual command shorthands
								if(!j.getAnnotation(Command.class).shorthand().isEmpty())
									if(!shortList.containsKey(j.getAnnotation(Command.class).shorthand()))
										shortList.put(j.getAnnotation(Command.class).shorthand(), id);
									else
										logger.warning("Conflicting shorthand for command " + id);
							}
							else
								logger.warning("Conflicting command " + id);
						else
							logger.warning("Command [" + id + "]: incorrect function prototype, thus won't be added to the command list.");
					}

		logger.info("... Done");
	}

	/**
	 * Parses a String into a primitive variable. Choice between boolean, byte, short, int, long, float, double, string
	 * @param target Target primitive
	 * @param value Input value
	 * @return `null` if the value is null or the string content is "null", or the parsed value.
	 * @throws IllegalArgumentException If the target primitive/class isn't part of the previous list
	 */
	public static Object parse(Class<?> target, String value) throws IllegalArgumentException
	{
		if(value         == null   || value.equals("null")  ) return null;
		if(Boolean.class == target || Boolean.TYPE == target) return Boolean.parseBoolean(value);
		if(Byte.class    == target || Byte.TYPE    == target) return Byte.parseByte(value);
		if(Short.class   == target || Short.TYPE   == target) return Short.parseShort(value);
		if(Integer.class == target || Integer.TYPE == target) return Integer.parseInt(value);
		if(Long.class    == target || Long.TYPE    == target) return Long.parseLong(value);
		if(Float.class   == target || Float.TYPE   == target) return Float.parseFloat(value);
		if(Double.class  == target || Double.TYPE  == target) return Double.parseDouble(value);
		if(String.class  == target)                           return value;
		throw new IllegalArgumentException("Unknown target type");
	}

	/**
	 * Gets the command list (cloned to avoid modifications)
	 * @return The current cloned command list
	 */
	public HashMap<String, InvokableMethod> getCommandList()
	{
		return (HashMap<String, InvokableMethod>) commandList.clone();
	}

	/**
	 * Gets the shorthands list (cloned to avoid edits)
	 * @return The current cloned short list
	 */
	public HashMap<String, String> getShortList()
	{
		return (HashMap<String, String>) shortList.clone();
	}
}
