package net.pingex.discordbot;

import sx.blah.discord.api.ClientBuilder;
import sx.blah.discord.api.IDiscordClient;
import sx.blah.discord.api.IListener;
import sx.blah.discord.handle.impl.events.ReadyEvent;
import sx.blah.discord.util.DiscordException;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;

/**
 * Main Entry Point
 * @version 0.1-dev
 * @author Raphael "Pingex" NAAS
 */
class DiscordBot
{
	private static final Logger LOGGER = Logger.getLogger(DiscordBot.class.getName());

	/**
	 * Discord Client instance
	 */
	private IDiscordClient client;

	/**
	 * Constructor of the singleton
	 * @param client Discord client instance used to Bootstrap the whole program
	 */
	private DiscordBot(IDiscordClient client)
	{
		this.client = client;
		LOGGER.info("Successfully logged in !");

		this.client.getDispatcher().registerListener((IListener<ReadyEvent>) event ->
				LOGGER.info("Logged in as \"" + event.getClient().getOurUser().getName() + "#" + event.getClient().getOurUser().getDiscriminator() + "\" (#" + event.getClient().getOurUser().getID() + ")"));

		enableModules();
		this.client.getDispatcher().registerListener(CommandDispatcher.getInstance());
	}

	private void enableModules()
	{
		LOGGER.info("Enabling modules");

		// Hardcoded internal modules
		new InternalCommandsModule(client);

		if(!Configuration.exists("plugins", "enable") || Configuration.getValue("plugins", "enable").isEmpty())
		{
			LOGGER.warning("Key \"plugins/enable\" doesn't exist in Configuration. No module will be enabled.");
			return;
		}

		for(String i : Configuration.getValue("plugins", "enable").split(","))
		{
			i = i.trim();
			try
			{
				Class<?> clazz = Class.forName(i);
				if(!AbstractModule.class.isAssignableFrom(clazz))
					throw new InvalidModuleException(i + " isn't a valid module.");
				clazz.getConstructor(IDiscordClient.class).newInstance(client);
			} catch (ClassNotFoundException e)
			{
				LOGGER.warning("Class " + i + " wasn't found, thus won't be loaded.");
			} catch (InvalidModuleException e)
			{
				LOGGER.warning(e.getMessage());
			} catch (NoSuchMethodException e)
			{
				LOGGER.warning("No valid Constructor for " + i + " was found.");
			} catch (InstantiationException | IllegalAccessException | InvocationTargetException e)
			{
				LOGGER.warning("An error occured while creating an instance of " + i);
				e.printStackTrace();
			}
		}
	}

	// ====================================================
	// Static functions: main, config loader, etc.
	// ====================================================

	/**
	 * Entry point
	 * @param args Arguments passed to java
	 */
	public static void main(String[] args)
	{
		// SET UP
		System.setProperty("java.util.logging.SimpleFormatter.format", "%1$tY-%1$tm-%1$td %1$tH:%1$tM:%1$tS [%2$s/%4$s] %5$s%6$s%n");

		LOGGER.info("Hello World, starting up");
		LOGGER.info("DiscordBot version " + (DiscordBot.class.getPackage().getImplementationVersion() != null ? DiscordBot.class.getPackage().getImplementationVersion() : "UNKNOWN"));

		// LOAD CONFIGURATION
		try
		{
			Configuration.loadConfiguration("config.ini"); // TODO: pass configuration file using commandline argument
		} catch (IOException e)
		{
			LOGGER.severe("Could not load Configuration, reason: " + e.getMessage());
			System.exit(10);
		}

		LOGGER.info("Logging in to Discord");
		try
		{
			loginToDiscord();
		} catch (ConfigurationException e)
		{
			LOGGER.severe("There is a problem with the Configuration: " + e.getMessage());
			System.exit(9);
		} catch (DiscordException e)
		{
			LOGGER.severe(e.getMessage());
			System.exit(2);
		}
	}

	/**
	 * Discord login subroutine. Called by `main()` once
	 * @return Instance of DiscordBot if login was successful
	 * @throws ConfigurationException When token/email/password field in config.ini isn't filled
	 * @throws DiscordException
	 */
	private static DiscordBot loginToDiscord() throws ConfigurationException, DiscordException
	{
		ClientBuilder builder = new ClientBuilder();
		if(Configuration.getValue("discord", "isBot", boolean.class))
		{
			if(!Configuration.exists("discord", "token"))
				throw new ConfigurationException("No login token defined");
			builder.withToken(Configuration.getValue("discord", "token"));
		}
		else
		{
			if(!Configuration.exists("discord", "email") || !Configuration.exists("discord", "password"))
				throw new ConfigurationException("Can't login: Either the email or the password wasn't provided.");
			builder.withLogin(Configuration.getValue("discord", "email"), Configuration.getValue("discord", "password"));
		}

		return new DiscordBot(builder.login());
	}
}
